package de.hs_mannheim.informatik.spreadsheet;

/**
 * Part of a simplified spreadsheet system for the PR1 programming lab at Hochschule Mannheim.
 * A cell needs to be able to hold a formula and a value 
 *
 * @author  Oliver Hummel & Victor Waitz
 */
public class Cell {
    private String formula = "";
    private String value = "";

    /**
     * Public method to get the formula of the cell.
     * @return formula of the cell.
     */
    public String getFormula() {
        return formula;
    }

    /**
     * Public method to set the formula of the cell.
     * @param formula formula of the cell.
     */
    public void setFormula(String formula) {
        if (!formula.isEmpty())
            this.formula = formula.toUpperCase().substring(1);  // removes '=' at the beginning
    }

    /**
     * Public method to get the value of the cell.
     * @return value of the cell
     */
    public String getValue() {
        return value;
    }

    /**
     * Public method to set the value of the cell.
     * @param value value of the cell.
     */
    public void setValue(String value) {
        this.value = value;
    }

    /**
     * Public method to get cell value as a string.
     * @return value as string.
     */
    public String toString() {
        return String.format("%4s", value);
    }

    /**
     * Public method to check if the cell is empty.
     * @return true if the cell is empty, false otherwise.
     */
    public boolean isEmpty() {
        return value.isEmpty();
    }
    
}
